﻿using System;
using System.IO;
using System.Data.OleDb;
using System.Data.SqlClient;
using System.Reflection;
using System.Runtime.InteropServices;

[assembly: AssemblyTitle("oledb-sql")]
[assembly: AssemblyProduct("oledb-sql")]
[assembly: AssemblyCopyright("Copyright © 2011-2019 Sergey Vaselenko")]

[assembly: ComVisible(false)]
[assembly: AssemblyVersion("1.1.0.0")]
[assembly: AssemblyFileVersion("1.1.0.0")]

namespace oledb_sql
{
    class Program
    {

        /// <summary>
        /// Expands a full path for a connection string data source.
        /// </summary>
        /// <param name="connectionString">The connection string</param>
        /// <returns>
        /// The connection string with a full path of a data source.
        /// </returns>
        /// <remarks>
        /// The OleDB standard dictates a use of a full path for a data source.
        /// </remarks>
        static string ExpandConnectionStringFileName(string connectionString)
        {
            const string term = "Data Source=";

            int index = connectionString.IndexOf(term);
            if (index < 0)
                return connectionString;

            string dataSource = connectionString.Substring(index + term.Length,
                connectionString.IndexOf(";", index + term.Length) - index - term.Length);

            if (!File.Exists(dataSource))
                return connectionString;

            string expandedDataSource = Path.GetFullPath(dataSource);

            // Console.WriteLine("DataSource expanded to {0}", expandedDataSource);

            connectionString.Replace(term + dataSource, term + expandedDataSource);

            return connectionString;

        }

        /// <summary>
        /// Executes BulkCopy for OleDB data to SQL Server
        /// </summary>
        /// <param name="oleDbConnectionString">The OleDB connection string</param>
        /// <param name="oleDbSQL">The SQL code to retrive OleDB data</param>
        /// <param name="sqlConnectionString">The SQL Server connection string</param>
        /// <param name="sqlTableName">The SQL Server destination table</param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Security",
            "CA2100:Review SQL queries for security vulnerabilities")]
        static void OleDBtoSqlServer(string oleDbConnectionString, string oleDbSQL,
            string sqlConnectionString, string sqlTableName)
        {

            oleDbConnectionString = ExpandConnectionStringFileName(oleDbConnectionString);

            OleDbConnection oleDbConnection = new OleDbConnection(oleDbConnectionString);
            try
            {
                oleDbConnection.Open();
                OleDbCommand command = new OleDbCommand(oleDbSQL, oleDbConnection);
                try
                {
                    OleDbDataReader reader = command.ExecuteReader();
                    try
                    {
                        SqlConnection sqlConnection = new SqlConnection(sqlConnectionString);
                        try
                        {
                            sqlConnection.Open();
                            using (SqlBulkCopy bulkCopy = new SqlBulkCopy(sqlConnection))
                            {
                                bulkCopy.DestinationTableName = sqlTableName;
                                bulkCopy.WriteToServer(reader);
                            }
                        }
                        finally
                        {
                            sqlConnection.Close();
                        }
                    }
                    finally
                    {
                        reader.Close();
                    }
                }
                finally
                {
                    command.Dispose();
                }
            }
            finally
            {
                oleDbConnection.Close();
            }
        }

        /// <summary>
        /// Formats value for CSV output
        /// </summary>
        /// <param name="value">The value to output</param>
        /// <param name="delimiter">The CSV column delimiter</param>
        /// <returns>
        /// The formatted value.
        /// </returns>
        /// <remarks>
        /// Function qoutes string values which contains delimiter or qoutes.
        /// </remarks>
        static Object FormatValueForCSV(Object value, string delimiter)
        {
            switch (value.GetType().Name)
            {
                case "String":
                    if (value.ToString().IndexOf('"') >= 0)
                    {
                        return '"' + value.ToString().Replace(@"""", @"""""") + '"';
                    }
                    else if (value.ToString().IndexOf(delimiter) >= 0)
                    {
                        return '"' + value.ToString() + '"';
                    }
                    else
                        return value.ToString();
                default:
                    return value;
            }

        }

        /// <summary>
        /// Outputs OleDB data to the console using CSV format
        /// </summary>
        /// <param name="oleDbConnectionString">The OleDB connection string</param>
        /// <param name="oleDbSQL">The SQL code to retrive OleDB data</param>
        /// <param name="delimiter">The CSV column delimiter</param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Security",
            "CA2100:Review SQL queries for security vulnerabilities")]
        static void OleDBtoCSV(string oleDbConnectionString, string oleDbSQL, string delimiter)
        {
            oleDbConnectionString = ExpandConnectionStringFileName(oleDbConnectionString);

            OleDbConnection oleDbConnection = new OleDbConnection(oleDbConnectionString);
            try
            {
                oleDbConnection.Open();
                OleDbCommand command = new OleDbCommand(oleDbSQL, oleDbConnection);
                try
                {
                    OleDbDataReader reader = command.ExecuteReader();
                    try
                    {
                        Console.Write("{0}", reader.GetName(0));
                        for (int i = 1; i < reader.FieldCount; i++)
                        {
                            Console.Write("{1}{0}", reader.GetName(i), delimiter);
                        }
                        Console.WriteLine();

                        while (reader.Read())
                        {
                            Console.Write("{0}", reader.GetValue(0));
                            for (int i = 1; i < reader.FieldCount; i++)
                            {
                                Console.Write("{1}{0}",
                                    FormatValueForCSV(reader.GetValue(i), delimiter), delimiter);
                            }
                            Console.WriteLine();
                        }
                    }
                    finally
                    {
                        reader.Close();
                    }
                }
                finally
                {
                    command.Dispose();
                }
            }
            finally
            {
                oleDbConnection.Close();
            }
        }

        /// <summary>
        /// Outputs Help to the console
        /// </summary>
        static void Help()
        {
            PortableExecutableKinds peKind;
            ImageFileMachine machine;
            Assembly.GetExecutingAssembly().GetModules()[0].GetPEKind(out peKind, out machine);

            string platform;
            if (machine == ImageFileMachine.AMD64)
                platform = "64-bit";
            else if ((peKind & PortableExecutableKinds.Required32Bit) > 0)
                platform = "32-bit";
            else
                platform = "      ";

            Console.WriteLine("oledb-sql 1.1    The command-line utility for copying OleDB data to SQL Server");
            Console.WriteLine("{0}           Copyright 2011-2019 Sergey Vaselenko, www.excel-sql-server.com", platform);
            Console.WriteLine();
            Console.WriteLine("Usage: oledb-sql sql/csv");
            Console.WriteLine("  sql <OleDB con string> <OleDB SQL> <SQL Server con string> <SQL Server table>");
            Console.WriteLine("  csv <OleDB con string> <OleDB SQL> <Delimiter>");
            Console.WriteLine();
            Console.WriteLine("Microsoft Excel 2003/2007/2010/2013/2016/2019 files can be used as OleDB data source.");
            Console.WriteLine("See examples at https://www.excel-sql-server.com");
            Console.WriteLine();
            Console.WriteLine("For batch operations you may create task files with parameters, for example");
            Console.WriteLine("oledb-sql-task-01.txt, for Excel 2003 workbook test.xls like this:");
            Console.WriteLine(@"OleDbConString==Provider=Microsoft.Jet.OLEDB.4.0;Data Source=test.xls;Extended Properties='Excel 8.0';");
            Console.WriteLine(@"OleDbSQL==SELECT * FROM [Sheet1$]");
            Console.WriteLine(@"SqlConString==Data Source=.;Initial Catalog=TestDB;Integrated Security=True;");
            Console.WriteLine(@"SqlTable==dbo.ExcelImport");
            Console.WriteLine(@"CsvFile==ExcelImport1.csv");
            Console.WriteLine();
            Console.WriteLine("or oledb-sql-task-02.txt, for Excel 2007/2010/2013/2016/2019 workbook test.xlsx like this:");
            Console.WriteLine(@"OleDbConString==Provider=Microsoft.ACE.OLEDB.12.0;Data Source=test.xlsx;Extended Properties='Excel 12.0';");
            Console.WriteLine(@"OleDbSQL==SELECT * FROM [Sheet1$]");
            Console.WriteLine(@"SqlConString==Data Source=.;Initial Catalog=TestDB;Integrated Security=True;");
            Console.WriteLine(@"SqlTable==dbo.ExcelImport");
            Console.WriteLine(@"CsvFile==ExcelImport2.csv");
            Console.WriteLine();
            Console.WriteLine("Then use a command file, for example oledb-sql-task-run-sql.cmd, like this:");
            Console.WriteLine("@echo off");
            Console.WriteLine("for %%i in (oledb-sql-task-*.txt) do call :RUN_ONE %%i");
            Console.WriteLine("goto END");
            Console.WriteLine(":RUN_ONE");
            Console.WriteLine(@"for /F ""eol=; tokens=1* delims==="" %%i in (%1) do set %%i=%%j");
            Console.WriteLine(@"oledb-sql.exe sql ""%OleDbConString%"" ""%OleDbSQL%"" ""%SqlConString%"" ""%SqlTable%""");
            Console.WriteLine(@"rem oledb-sql.exe csv ""%OleDbConString%"" ""%OleDbSQL%"" , > %CsvFile%");
            Console.WriteLine(":END");
            Console.WriteLine();
            Environment.ExitCode = 1;
        }

        /// <summary>
        /// The command-line utility for copying OleDB data to SQL Server
        /// </summary>
        /// <param name="args">
        /// <list type="bullet">
        /// <item>sql - copies OleDB data to SQL Server</item>
        /// <item>csv - outputs OleDB data to the console using CSV format</item>
        /// </list>
        /// </param>
        static void Main(string[] args)
        {
            try
            {
                if (args.Length == 0)
                {
                    Help();
                    return;
                }

                switch (args[0].ToLower())
                {
                    case "sql":
                        if (args.Length != 5)
                        {
                            Help();
                            break;
                        }
                        OleDBtoSqlServer(args[1], args[2], args[3], args[4]);
                        break;
                    case "csv":
                        if (args.Length != 4)
                        {
                            Help();
                            break;
                        }
                        OleDBtoCSV(args[1], args[2], args[3]);
                        break;
                    default:
                        Help();
                        break;
                }
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
                Environment.ExitCode = 2;
            }
        }
    }
}
